@@ -22,16 +22,32 @@ seq2 <- function(to, by)
 ### bplapply2()
 ###
 ### A simple wrapper to BiocParallel::bplapply() that falls back to lapply()
-### if 'BPPARAM' is NULL.
+### if 'length(X)' <= 1 or 'BPPARAM' is NULL or 'bpnworkers(BPPARAM)' <= 1.
+###
+### Also disable HDF5 file locking for the duration of the bplapply() loop,
+### just in case the user is dealing with HDF5-backed DelayedArray objects
+### or other objects that involve HDF5 files. This is needed for concurrent
+### access to a single dataset from multiple processes, even for read-only
+### access!
+### See https://portal.hdfgroup.org/display/knowledge/Questions+about+thread-safety+and+concurrent+access
 ###
 
 bplapply2 <- function(X, FUN, ..., BPPARAM=NULL)
 {
-    if (is.null(BPPARAM))
+    if (length(X) <= 1L || is.null(BPPARAM))
+        return(lapply(X, FUN, ...))
+    S4Vectors:::load_package_gracefully("BiocParallel", "for this operation")
+    ## The only value that is allowed for HDF5_USE_FILE_LOCKING is "FALSE"
+    ## (all uppercase).
+    ## Note sure this is actually a good idea so commenting out for now.
+    #if (Sys.getenv("HDF5_USE_FILE_LOCKING") != "FALSE") {
+    #    Sys.setenv(HDF5_USE_FILE_LOCKING="FALSE")
+    #    on.exit(Sys.unsetenv("HDF5_USE_FILE_LOCKING"))
+    #}
+    if (!is(BPPARAM, "BiocParallelParam"))
+        stop(wmsg("'BPPARAM' must be a BiocParallelParam derivative"))
+    if (BiocParallel::bpnworkers(BPPARAM) <= 1L)
         return(lapply(X, FUN, ...))
-    if (!requireNamespace("BiocParallel", quietly=TRUE))
-        stop(wmsg("Couldn't load the BiocParallel package. Please ",
-                  "install the BiocParallel package and try again."))
     BiocParallel::bplapply(X, FUN, ..., BPPARAM=BPPARAM)
 }
 
@@ -87,29 +103,33 @@ LCM <- function(x, y)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### normarg_dim() and normarg_dimnames()
+### as_integer_dim(), normarg_dim() and normarg_dimnames()
 ###
 
-normarg_dim <- function(dim, argname="dim")
+as_integer_dim <- function(dim, what)
 {
-    if (!is.numeric(dim))
-        stop(wmsg("'", argname, "' must be an integer vector"))
+    if (!(is.numeric(dim) && is.vector(dim)))
+        stop(wmsg(what, " must be an integer vector"))
     if (length(dim) == 0L)
-        stop(wmsg("'", argname, "' cannot be an empty vector"))
+        stop(wmsg(what, " cannot be an empty vector"))
     if (any(is.na(dim)))
-        stop(wmsg("'", argname, "' cannot contain NAs"))
+        stop(wmsg(what, " cannot contain NAs"))
     if (any(dim < 0))
-        stop(wmsg("'", argname, "' cannot contain negative values"))
+        stop(wmsg(what, " cannot contain negative values"))
     if (!is.integer(dim)) {
         if (any(dim > .Machine$integer.max))
-            stop(wmsg("'", argname, "' cannot contain values greater ",
-                      "than '.Machine$integer.max' (= 2^31-1 = ",
-                      .Machine$integer.max, ")"))
+            stop(wmsg(what, " cannot contain values greater ",
+                      "than 2^31-1 (= .Machine$integer.max)"))
         dim <- as.integer(dim)
     }
     dim
 }
 
+normarg_dim <- function(dim, argname="dim")
+{
+    as_integer_dim(dim, paste0("'", argname, "'"))
+}
+
 normarg_dimnames <- function(dimnames, dim)
 {
     ndim <- length(dim)
@@ -144,22 +164,45 @@ normarg_dimnames <- function(dimnames, dim)
 }
 
 
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Right-pad dim vector with ones
+###
+
+rpad_dim <- function(dim, ndim)
+{
+    stopifnot(is.integer(dim), isSingleInteger(ndim))
+    nd <- ndim - length(dim)
+    if (nd > 0L)
+        dim <- c(dim, rep.int(1L, nd))
+    dim
+}
+
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### 2 wrappers to dim<- and dimnames<- that try to avoid unnecessary copies
 ### of 'x'
 ###
 
-### Preserves the dimnames when dropping or adding rightmost dimensions (i.e.
-### outermost or slowest moving dimensions in the memory layout).
+### Preserves the dimnames when dropping or adding **rightmost dimensions**
+### (a.k.a. outermost or slowest moving dimensions in the memory layout).
 set_dim <- function(x, value)
 {
-    stopifnot(is.numeric(value))
-    if (!is.integer(value))
-        value <- as.integer(value)
+    if (is.numeric(value)) {
+        if (!is.integer(value))
+            value <- as.integer(value)
+    } else {
+        stopifnot(is.null(value))
+    }
     x_dim <- dim(x)
     if (identical(x_dim, value))
         return(x)
-    x_dimnames <- dimnames(x)
+
+    if (is.null(x_dim)) {
+        ## Treat vector as 1D array.
+        x <- as.array(x)
+        x_dim <- dim(x)
+    }
+    x_dimnames <- simplify_NULL_dimnames(dimnames(x))
     dim(x) <- value
     if (!is.null(x_dimnames)) {
         old_ndim <- length(x_dim)
@@ -180,6 +223,21 @@ set_dimnames <- function(x, value)
     x
 }
 
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### set_or_check_dim()
+###
+
+set_or_check_dim <- function(x, dim)
+{
+    x_dim <- dim(x)
+    if (is.null(x_dim)) {
+        dim(x) <- dim
+    } else {
+        stopifnot(identical(x_dim, dim))
+    }
+    x
+}
+
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### get_first_non_NULL_dimnames()
@@ -303,22 +361,6 @@ validate_dimnames_slot <- function(x, dim, slotname="dimnames")
 }
 
 
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### set_or_check_dim()
-###
-
-set_or_check_dim <- function(x, dim)
-{
-    x_dim <- dim(x)
-    if (is.null(x_dim)) {
-        dim(x) <- dim
-    } else {
-        stopifnot(identical(x_dim, dim))
-    }
-    x
-}
-
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Zero-filled arrays
 ###
